/*
 * @Author: your name
 * @Date: 2021-11-16 09:18:58
 * @LastEditTime: 2021-12-16 11:10:42
 * @LastEditors: Please set LastEditors
 * @Description: In User Settings Edit
 * @FilePath: /roadmodel/src/CtApRoadModel/source/road/obstacle.h
 */
#pragma once
#include "obstacle.h"

namespace rm {
const int Obstacle::max_traj_points_ = 65;

void Obstacle::set_id(uint32_t id) { id_ = id; }

void Obstacle::set_exist_probability(float exist_probability) {
  exist_probability_ = exist_probability;
}

void Obstacle::set_classification(ObstacleType classification) {
  classification_ = classification;
}

void Obstacle::set_position(Vec2d position) { position_ = position; }

void Obstacle::set_center(Vec2d center) { center_ = center; }

void Obstacle::set_length(float length) { length_ = length; }

void Obstacle::set_width(float width) { width_ = width; }

// void Obstacle::set_heading(float heading) { heading_ = heading; }

void Obstacle::set_heading(float orientation) {
  if (motion_property_ == ObstacleMotionProperty::STATIONARY ||
      motion_property_ == ObstacleMotionProperty::STOPPED) {
    heading_ = orientation;
  } else {
    heading_ = atan2(speed_.y(), speed_.x());
  }
}

void Obstacle::set_yaw_rate(float yaw_rate) { yaw_rate_ = yaw_rate; }

void Obstacle::set_speed(Vec2d speed) { speed_ = speed; }

void Obstacle::set_acceleration(Vec2d acceleration) {
  acceleration_ = acceleration;
}

void Obstacle::set_kappa(float kappa) { kappa_ = kappa; }

void Obstacle::set_moving_direction(uint8_t movingdirection) {
  moving_direction_ = ObstacleMovingDirection(movingdirection);
}

void Obstacle::set_motion_property(uint8_t motion_property) {
  motion_property_ = ObstacleMotionProperty(motion_property);
}

void Obstacle::set_is_front_obstacle(bool is_front_obstacle) {
  is_front_obstacle_ = is_front_obstacle;
}

void Obstacle::Set_existence() { existence_flag_ = true; }

void Obstacle::Unset_existence() {
  existence_flag_ = false;
  trplontype_changed_ = false;
  is_front_adjacent_target_ = false;
}

uint32_t Obstacle::id() const { return id_; }

float Obstacle::exist_probability() const { return exist_probability_; }

ObstacleType Obstacle::classification() const { return classification_; }

ObstacleMotionProperty Obstacle::motion_property() const {
  return motion_property_;
}

ObstacleMovingDirection Obstacle::moving_direction() const {
  return moving_direction_;
}

Vec2d Obstacle::position() const { return position_; }

Vec2d Obstacle::center() const { return center_; }

float Obstacle::length() const { return length_; }

float Obstacle::width() const { return width_; }

float Obstacle::heading() const { return heading_; }

float Obstacle::yaw_rate() const { return yaw_rate_; }

Vec2d Obstacle::speed() const { return speed_; }

Vec2d Obstacle::acceleration() const { return acceleration_; }

float Obstacle::kappa() const { return kappa_; }

bool Obstacle::existence() const { return existence_flag_; }

const std::list<Vec2d> &Obstacle::historical_traj() const {
  return historical_traj_;
}

bool Obstacle::delayed_motion() const { return delayed_motion_; }

bool Obstacle::is_front_obstacle() const { return is_front_obstacle_; }

ReferencePointType Obstacle::reference_point_type() const {
  return reference_point_type_;
}

void Obstacle::set_reference_point_type(
    ReferencePointType reference_point_type) {
  reference_point_type_ = reference_point_type;
}

std::ostream &operator<<(std::ostream &os, const Obstacle &obstacle) {
  os << "id = " << obstacle.id() << ", x = " << obstacle.center().x()
     << ", y = " << obstacle.center().y() << ", h = " << obstacle.heading()
     << ", v = " << obstacle.speed().Length()
     << ", a = " << obstacle.acceleration().Length()
     << ", cur = " << obstacle.kappa();
  return os;
}

// update the history pos point according to ego vehicle motion
void Obstacle::Update_pos(double angle, double dx) {
  for (auto &&it : historical_traj_) {
    it.SelfRotate(-angle); // rotate
    it.set_x(it.x() - dx); // translate
  }
}

void Obstacle::Record_pos_center(Vec2d newpos, Vec2d newcenter) {
  // only consider 50 recent position
  set_position(newpos);
  set_center(newcenter);
  // DLOG(INFO) << PRINT(newcenter.x());
  // DLOG(INFO) << PRINT(filtered_state_.at(0));
  // DLOG(INFO) << PRINT(newcenter.y());
  // DLOG(INFO) << PRINT(filtered_state_.at(1));
  Vec2d filtered_center = {filtered_state_.at(0), filtered_state_.at(1)};
  if (!historical_traj_.empty()) {
    if (historical_traj_.back().DistanceTo(filtered_center) >
        0.8) // the distance between traj point shall be greater than 1.0 m
    {
      historical_traj_.push_back(filtered_center);
      if (historical_traj_.size() > max_traj_points_) {
        historical_traj_.pop_front();
      } else {
        // do nothing
      }
    } else {
      // do nothing
    }
  } else {
    historical_traj_.push_back(filtered_center);
  }
}

void Obstacle::Set_kappa() {
  if (historical_traj_.size() >= 3) {
    Eigen::Matrix3d Mat_Cir, Mat_Inv;
    Eigen::Vector3d Vec_Cir, Para_Cir;
    std::list<Vec2d>::iterator obj_it = historical_traj_.end();
    std::array<Vec2d, 3> objpos_vec;
    bool Mat_invertible;

    for (int i = 0; i < 3; i++) {
      obj_it--;
      objpos_vec[i] = *obj_it;
    }
    Mat_Cir << 2 * objpos_vec[0].x(), 2 * objpos_vec[0].y(), 1,
        2 * objpos_vec[1].x(), 2 * objpos_vec[1].y(), 1, 2 * objpos_vec[2].x(),
        2 * objpos_vec[2].y(), 1;
    Vec_Cir << -(objpos_vec[0].x() * objpos_vec[0].x() +
                 objpos_vec[0].y() * objpos_vec[0].y()),
        -(objpos_vec[1].x() * objpos_vec[1].x() +
          objpos_vec[1].y() * objpos_vec[1].y()),
        -(objpos_vec[2].x() * objpos_vec[2].x() +
          objpos_vec[2].y() * objpos_vec[2].y());
    Mat_Cir.computeInverseWithCheck(Mat_Inv, Mat_invertible);

    if (Mat_invertible) {
      Para_Cir = Mat_Inv * Vec_Cir;
      kappa_ = 1 / sqrt(Para_Cir(0) * Para_Cir(0) + Para_Cir(1) * Para_Cir(1) -
                        Para_Cir(2));
    } else {
      kappa_ = 0.0;
    }
  } else {
    kappa_ = 0.0;
  }
}

// const OneDimLookupTable Obstacle::in_corridor_long_dist_inner_table_ =
//     OneDimLookupTable(
//         {-100, -20, -15, -10, 0, 5.0, 10, 15, 20, 30, 40, 50, 60, 80, 100},
//         {1.0, 1.0, 1.0, 1.5, 2.0, 1.1, 1.0, 1.0, 1.0, 1.0, 1.0, 1.5, 1.5, 4.0,
//          4.0}); // factor
// const OneDimLookupTable Obstacle::in_corridor_long_dist_outer_table_ =
//     OneDimLookupTable(
//         {-100, -20, -15, -10, 10, 15, 20, 30, 40, 50},
//         {1.0, 1.0, 1.0, 1.0, 0.8, 0.8, 0.9, 1.0, 1.0, 1.0}); // factor
// const OneDimLookupTable Obstacle::in_corridor_inner_width_table_ =
//     OneDimLookupTable({2.4, 2.8, 3.2, 3.6, 4.0, 4.4},
//                       {-0.1, -0.1, -0.1, -0.2, -0.3, -0.35}); // base
// const OneDimLookupTable Obstacle::in_corridor_outer_width_table_ =
//     OneDimLookupTable({2.4, 2.8, 3.2, 3.6, 4.0, 4.4},
//                       {0.1, 0.1, 0.1, 0.1, 0.1, 0.2});
// const OneDimLookupTable Obstacle::in_corridor_ego_position_left_table_ =
//     OneDimLookupTable({-1.2, -1.0, -0.8, -0.2}, {1.2, 1.0, 0.8, 0.0});
// const OneDimLookupTable Obstacle::in_corridor_ego_position_right_table_ =
//     OneDimLookupTable({0.2, 0.8, 1.0, 1.2}, {0.0, -0.8, -1.0, -1.2});
// const OneDimLookupTable Obstacle::in_corridor_ego_position_table_ =
//     OneDimLookupTable({-1.2, -1.0, -0.8, -0.2, 0.2, 0.8, 1.0, 1.2},
//                       {0.5, 0.4, 0.2, 0.0, 0.0, -0.2, -0.4, -0.5});
// const OneDimLookupTable Obstacle::width_factor_dl_car_ =
//     OneDimLookupTable({0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 1.0},
//                       {0.0, 0.2, 0.3, 0.5, 0.5, 0.5, 0.8, 1.2, 1.2, 1.5});
// const OneDimLookupTable Obstacle::width_factor_dl_trunck_ =
//     OneDimLookupTable({0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 1.0},
//                       {0.0, 0.5, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.4, 1.5});
OneDimLookupTable Obstacle::in_corridor_long_dist_inner_table_ =
    OneDimLookupTable(
        GetRMConfig().in_corridor_long_dist_inner_table_x,
        GetRMConfig().in_corridor_long_dist_inner_table_y); // factor
OneDimLookupTable Obstacle::in_corridor_long_dist_outer_table_ =
    OneDimLookupTable(
        GetRMConfig().in_corridor_long_dist_outer_table_x,
        GetRMConfig().in_corridor_long_dist_outer_table_y); // factor
OneDimLookupTable Obstacle::in_corridor_inner_width_table_ =
    OneDimLookupTable(GetRMConfig().in_corridor_inner_width_table_x,
                      GetRMConfig().in_corridor_inner_width_table_y); // base
OneDimLookupTable Obstacle::in_corridor_outer_width_table_ =
    OneDimLookupTable(GetRMConfig().in_corridor_outer_width_table_x,
                      GetRMConfig().in_corridor_outer_width_table_y);
OneDimLookupTable Obstacle::in_corridor_ego_position_left_table_ =
    OneDimLookupTable(GetRMConfig().in_corridor_ego_position_left_table_x,
                      GetRMConfig().in_corridor_ego_position_left_table_y);
OneDimLookupTable Obstacle::in_corridor_ego_position_right_table_ =
    OneDimLookupTable(GetRMConfig().in_corridor_ego_position_right_table_x,
                      GetRMConfig().in_corridor_ego_position_right_table_y);
OneDimLookupTable Obstacle::in_corridor_ego_position_table_ =
    OneDimLookupTable(GetRMConfig().in_corridor_ego_position_table_x,
                      GetRMConfig().in_corridor_ego_position_table_y);
OneDimLookupTable Obstacle::width_factor_dl_car_ = OneDimLookupTable(
    GetRMConfig().width_factor_dl_car_x, GetRMConfig().width_factor_dl_car_y);
OneDimLookupTable Obstacle::width_factor_dl_trunck_ =
    OneDimLookupTable(GetRMConfig().width_factor_dl_trunck_x,
                      GetRMConfig().width_factor_dl_trunck_y);
// const OneDimLookupTable
//     Obstacle::in_corridor_upper_long_dis_ego_velocity_table_ =
//         OneDimLookupTable({30, 50, 70, 80, 100, 120},
//                           {55, 75, 105, 115, 135, 155});
// const OneDimLookupTable
//     Obstacle::in_corridor_lower_long_dis_ego_velocity_table_ =
//         OneDimLookupTable({30, 50, 70, 80, 100, 120},
//                           {50, 70, 100, 110, 130, 150});
OneDimLookupTable Obstacle::in_corridor_upper_long_dis_ego_velocity_table_ =
    OneDimLookupTable(
        GetRMConfig().in_corridor_upper_long_dis_ego_velocity_table_x,
        GetRMConfig().in_corridor_upper_long_dis_ego_velocity_table_y);
OneDimLookupTable Obstacle::in_corridor_lower_long_dis_ego_velocity_table_ =
    OneDimLookupTable(
        GetRMConfig().in_corridor_lower_long_dis_ego_velocity_table_x,
        GetRMConfig().in_corridor_lower_long_dis_ego_velocity_table_y);
// const OneDimLookupTable Obstacle::width_factor_s_ = OneDimLookupTable(
//     {-1, 0, 1, 5.5, 6, 8, 10, 15, 20, 30, 40, 50, 60, 80, 100},
//     {1.0, 0, 0.2, 0.5, 0.9, 0.9, 0.95, 0.95, 0.95, 0.98, 0.98, 1.0, 1.0, 1.0,
//      1.0});
// const OneDimLookupTable Obstacle::width_factor_ego_yaw_rate_table_ =
//     OneDimLookupTable({0.0, 1.0, 2.0, 2.5, 3.0, 3.5, 4.0},
//                       {1.0, 0.85, 0.8, 0.8, 0.5, 0.2, 0.0});
OneDimLookupTable Obstacle::width_factor_s_ = OneDimLookupTable(
    GetRMConfig().width_factor_s_x, GetRMConfig().width_factor_s_y);
OneDimLookupTable Obstacle::width_factor_ego_yaw_rate_table_ =
    OneDimLookupTable(GetRMConfig().width_factor_ego_yaw_rate_table_x,
                      GetRMConfig().width_factor_ego_yaw_rate_table_y);
double Obstacle::width_factor_ego_yaw_rate_;
// bool Obstacle::In_other_corridor_check(const float lane_width,
//                                        const SLPoint &object) {
//   // object has not been assigned to any corridor
//   float half_lane_width = 0.5 * lane_width;
//   // double pred_time = prediction_time_dl_.Evaluate(std::abs(object.dl()));
//   double pred_time = 0.0;
//   double pred_obj_l =
//       object.l() + std::max(std::min(1.5, object.dl() * pred_time), -1.5);
//   DLOG(INFO) << PRINT(object.l());
//   DLOG(INFO) << PRINT(pred_obj_l);
//   bool is_in_corridor = std::abs(pred_obj_l) < half_lane_width;
//   return is_in_corridor;
// }

bool Obstacle::In_other_corridor_check(const Lane &lane,
                                       const Corridor *corridor,
                                       const SLPoint &object) {
  double half_lane_width = 0.5 * std::max(lane.lane_width(), 2.5f);
  // double pred_time = prediction_time_dl_.Evaluate(std::abs(object.dl()));
  double pred_time = 0.0;
  double pred_obj_l =
      object.l() + std::max(std::min(1.5, object.dl() * pred_time), -1.5);
  // DLOG(INFO) << PRINT(object.l());
  // DLOG(INFO) << PRINT(pred_obj_l);
  if ((lane.lane_position() == LanePosition::LEFT1 ||
       lane.lane_position() == LanePosition::RIGHT1) &&
      std::abs(object.s()) < 50.0) {
    if (std::abs(object.l() + 0.5 * width_) >
        std::abs(object.l() - 0.5 * width_)) {
      pred_obj_l = object.l() - 0.5 * width_ * 0.8;
    } else {
      pred_obj_l = object.l() + 0.5 * width_ * 0.8;
    }
  }
  bool is_in_corridor = std::abs(pred_obj_l) < half_lane_width;

  // following operations are needed to keep deboucner and delay from idling
  double inner_factor = in_corridor_long_dist_inner_table_.Evaluate(object.s());
  double outer_factor = in_corridor_long_dist_outer_table_.Evaluate(object.s());
  double corridor_width = corridor->drivable_width();
  double half_corridor_width = 0.5 * corridor_width;
  uint8_t lane_id = lane.lane_id();
  // left debouncer
  incorridor_debouncer_delay_map_[lane_id].first.left_debouncer.SetDebounce(
      half_corridor_width +
          in_corridor_inner_width_table_.Evaluate(corridor_width) *
              inner_factor,
      std::max(half_corridor_width +
                   in_corridor_outer_width_table_.Evaluate(corridor_width) *
                       outer_factor,
               half_lane_width));
  // right debouncer
  incorridor_debouncer_delay_map_[lane_id].first.right_debouncer.SetDebounce(
      std::min(-half_corridor_width -
                   in_corridor_outer_width_table_.Evaluate(corridor_width) *
                       outer_factor,
               -half_lane_width),
      -half_corridor_width -
          in_corridor_inner_width_table_.Evaluate(corridor_width) *
              inner_factor);
  bool left_debouncer_result_idle =
      incorridor_debouncer_delay_map_[lane_id]
          .first.left_debouncer.RunDebouceLessTure(pred_obj_l);
  bool right_debouncer_result_idle =
      incorridor_debouncer_delay_map_[lane_id]
          .first.right_debouncer.RunDebouceMoreTure(pred_obj_l);
  incorridor_debouncer_delay_map_[lane_id].first.in_corridor_delay.RunDelay(
      left_debouncer_result_idle && right_debouncer_result_idle);
  //

  return is_in_corridor;
}

bool Obstacle::In_ego_corridor_check(const Corridor *const corridor,
                                     const rm::Lane &lane,
                                     const SLPoint &object) {
  double inner_factor = in_corridor_long_dist_inner_table_.Evaluate(object.s());
  double outer_factor = in_corridor_long_dist_outer_table_.Evaluate(object.s());
  double corridor_width = corridor->drivable_width();
  double half_corridor_width = 0.5 * corridor_width;
  double half_lane_width = 0.5 * std::max(lane.lane_width(), 2.5f);
  uint8_t lane_id = lane.lane_id();
  // left debouncer
  incorridor_debouncer_delay_map_[lane_id].first.left_debouncer.SetDebounce(
      half_corridor_width +
          in_corridor_inner_width_table_.Evaluate(corridor_width) *
              inner_factor,
      std::max(half_corridor_width +
                   in_corridor_outer_width_table_.Evaluate(corridor_width) *
                       outer_factor,
               half_lane_width));
  // right debouncer
  incorridor_debouncer_delay_map_[lane_id].first.right_debouncer.SetDebounce(
      std::min(-half_corridor_width -
                   in_corridor_outer_width_table_.Evaluate(corridor_width) *
                       outer_factor,
               -half_lane_width),
      -half_corridor_width -
          in_corridor_inner_width_table_.Evaluate(corridor_width) *
              inner_factor);
  // double pred_time = prediction_time_dl_.Evaluate(std::abs(object.dl())) *
  //                    prediction_time_factor_s_.Evaluate(object.s());
  double pred_time = 0.0;
  double pred_obj_l =
      object.l() + std::max(std::min(1.5, object.dl() * pred_time), -1.5);
  // DLOG(INFO) << PRINT(object.l());
  // DLOG(INFO) << PRINT(pred_obj_l);
  bool left_debouncer_result =
      incorridor_debouncer_delay_map_[lane_id]
          .first.left_debouncer.RunDebouceLessTure(pred_obj_l);
  bool right_debouncer_result =
      incorridor_debouncer_delay_map_[lane_id]
          .first.right_debouncer.RunDebouceMoreTure(pred_obj_l);
  bool is_in_corridor =
      incorridor_debouncer_delay_map_[lane_id].first.in_corridor_delay.RunDelay(
          left_debouncer_result && right_debouncer_result);
  was_in_ego_corridor_assured_flfp_.Run(false, true);
  was_in_ego_corridor_assured_ = was_in_ego_corridor_assured_flfp_.get_status();
  return is_in_corridor;
}

bool Obstacle::In_ego_corridor_check(const Corridor *const corridor,
                                     const SLPoint &object, double ego_offset,
                                     const rm::Lane &lane,
                                     bool lane_changed_delay) {
  double inner_factor = in_corridor_long_dist_inner_table_.Evaluate(object.s());
  double outer_factor = in_corridor_long_dist_outer_table_.Evaluate(object.s());
  // DLOG(INFO) << PRINT(lane_changed_delay);
  // DLOG(INFO) << PRINT(ego_offset);
  double shift = lane_changed_delay ? 0.0
                                    : in_corridor_ego_position_table_.Evaluate(
                                          ego_offset); // deployed on left
  // DLOG(INFO) << PRINT(shift);
  double width = corridor->drivable_width();
  double half_width = std::max(std::min(0.5 * width, 2.5), 1.25);
  double half_lane_width = 0.5 * std::max(lane.lane_width(), 2.5f);
  // DLOG(INFO) << PRINT(half_lane_width);
  uint8_t lane_id = lane.lane_id();
  double object_l = object.l();
  bool is_iv_iiv_target =
      (trplontype_ == TargetType::IV || trplontype_ == TargetType::IIV);
  double pred_obj_l;
  if (classification_ == ObstacleType::PEDESTRIAN) {
    // left debouncer
    incorridor_debouncer_delay_map_[lane_id].first.left_debouncer.SetDebounce(
        half_lane_width, half_lane_width);
    // right debouncer
    incorridor_debouncer_delay_map_[lane_id].first.right_debouncer.SetDebounce(
        -half_lane_width, -half_lane_width);
    pred_obj_l = object_l;
  } else {
    double left_debouncer_outter =
               std::min(half_width +
                            in_corridor_outer_width_table_.Evaluate(width) *
                                outer_factor +
                            shift,
                        half_lane_width),
           right_debouncer_outter =
               std::max(-half_width -
                            in_corridor_outer_width_table_.Evaluate(width) *
                                outer_factor +
                            shift,
                        -half_lane_width);
    // left debouncer
    incorridor_debouncer_delay_map_[lane_id].first.left_debouncer.SetDebounce(
        std::min(half_width +
                     in_corridor_inner_width_table_.Evaluate(width) *
                         inner_factor +
                     shift,
                 half_lane_width),
        left_debouncer_outter);
    // right debouncer
    incorridor_debouncer_delay_map_[lane_id].first.right_debouncer.SetDebounce(
        right_debouncer_outter,
        std::max(-half_width -
                     in_corridor_inner_width_table_.Evaluate(width) *
                         inner_factor +
                     shift,
                 -half_lane_width));
    double object_closest_side;
    if (std::abs(object_l + 0.5 * width_) > std::abs(object_l - 0.5 * width_)) {
      object_closest_side = object_l - 0.5 * width_ * 0.8;
    } else {
      object_closest_side = object_l + 0.5 * width_ * 0.8;
    }
    bool is_in_ego_corridor_assured_ =
        (object_closest_side < left_debouncer_outter &&
         object_closest_side > right_debouncer_outter);
    was_in_ego_corridor_assured_flfp_.Run(
        is_in_ego_corridor_assured_ && is_iv_iiv_target, !is_iv_iiv_target);
    was_in_ego_corridor_assured_ =
        was_in_ego_corridor_assured_flfp_.get_status();
    double dl_width_factor = 0.0;
    if (object.s() <=
        std::max(
            50.0,
            std::min(2.5 * VehicleStateProvider::Instance()->lon_velocity(),
                     70.0)) && object.s() > 0.0) {
      // if current corridor is the first front one, then take into accout the
      // prediction time
      dl_width_factor =
          ((classification_ == ObstacleType::TRUCK)
               ? width_factor_dl_trunck_.Evaluate(std::abs(object.dl()))
               : width_factor_dl_car_.Evaluate(std::abs(object.dl()))) *
          width_factor_s_.Evaluate(object.s()) * width_factor_ego_yaw_rate_;
    }
    // DLOG(INFO) << PRINT(pred_time);
    // DLOG(INFO) << PRINT(object.dl());
    if (was_in_ego_corridor_assured_) {
      // if target is iv or iiv, then take into accout the object width
      // if (std::abs(object_l + 0.5 * width_) >
      //     std::abs(object_l - 0.5 * width_)) {
      //   pred_obj_l = object_l - 0.5 * width_ * 0.8;
      // } else {
      //   pred_obj_l = object_l + 0.5 * width_ * 0.8;
      // }
      pred_obj_l = object_closest_side;
      // DLOG(INFO) << PRINT(pred_obj_l);
    } else if (object.s() <=
               std::max(50.0, std::min(2.5 * VehicleStateProvider::Instance()
                                                 ->lon_velocity(),
                                       70.0))) {
      // consider the prediction
      if (std::abs(object_l + 0.5 * width_) >
          std::abs(object_l - 0.5 * width_)) {
        dl_width_factor = (object.dl() > 0.01) ? 0.5 : dl_width_factor;
        pred_obj_l = object_l - dl_width_factor * 0.5 * width_;
      } else {
        dl_width_factor = (object.dl() < -0.01) ? 0.5 : dl_width_factor;
        pred_obj_l = object_l + dl_width_factor * 0.5 * width_;
      }
    } else {
      pred_obj_l = object_l;
    }
    // DLOG(INFO) << PRINT(dl_width_factor);
  }
  // DLOG(INFO) << PRINT(object_l);
  // DLOG(INFO) << PRINT(pred_obj_l);
  bool left_debouncer_result =
      incorridor_debouncer_delay_map_[lane_id]
          .first.left_debouncer.RunDebouceLessTure(pred_obj_l);
  bool right_debouncer_result =
      incorridor_debouncer_delay_map_[lane_id]
          .first.right_debouncer.RunDebouceMoreTure(pred_obj_l);
  bool is_in_corridor =
      incorridor_debouncer_delay_map_[lane_id].first.in_corridor_delay.RunDelay(
          left_debouncer_result && right_debouncer_result);
  // DLOG(INFO) << PRINT(is_in_corridor);

  if (!is_in_corridor && (classification_ == ObstacleType::BICYCLE ||
                          classification_ == ObstacleType::PEDESTRIAN)) {
    vulnerable_object_critical_ego_corridor_debouncer_pair.excuted_flag = true;
    vulnerable_object_critical_ego_corridor_debouncer_pair.lat_dis_debouncer
        .SetDebounce(half_lane_width + 0.5, half_lane_width + 0.8);
    double long_lower_range =
        3.5 * std::max(VehicleStateProvider::Instance()->lon_velocity(), 2.0f);
    vulnerable_object_critical_ego_corridor_debouncer_pair.long_dis_debouncer
        .SetDebounce(long_lower_range, long_lower_range + 5.0);
    bool in_long_highlight_zone =
        vulnerable_object_critical_ego_corridor_debouncer_pair
            .long_dis_debouncer.RunDebouceLessTure(center_.x());
    bool in_lat_highlight_zone =
        vulnerable_object_critical_ego_corridor_debouncer_pair.lat_dis_debouncer
            .RunDebouceLessTure(std::abs(object_l));
    is_vulnerable_critical_obstacle_ =
        in_long_highlight_zone ? in_lat_highlight_zone : false;
  } else {
    is_vulnerable_critical_obstacle_ = false;
  }
  return is_in_corridor;
}

void Obstacle::Init_EKF(const std::vector<double> &measurement) {
  ekfilter_.Init(measurement);
  ekfilter_.GetState(filtered_state_);
}

void Obstacle::Run_EKF(const std::vector<double> &measurement) {
  ekfilter_.Predict();
  if (ekfilter_.Correct(measurement)) {
    ekfilter_.GetState(filtered_state_);
  } else {
    ekfilter_.Init(measurement);
    ekfilter_.GetState(filtered_state_);
  }
}

bool Obstacle::Run_oncoming_delay(const ObjectLocation &obj_loc,
                                  const uint8_t lane_id) {
  if (trplontype_ == TargetType::IV)
    return true;
  bool oncoming_status =
      (motion_property_ == ObstacleMotionProperty::MOVING ||
       motion_property_ == ObstacleMotionProperty::MOVING_SLOWLY)
          ? (moving_direction_ == ObstacleMovingDirection::ONCOMING)
          : std::abs(heading_) > 2.8;
  bool oncoming_flag =
      (std::abs(obj_loc.reference_point2corridor_frenet.ds()) < 1e-5)
          ? !oncoming_status
          : obj_loc.reference_point2corridor_frenet.ds() > 0.0 &&
                !oncoming_status;
  // oncoming_status =
  //     (classification_ == rm::ObstacleType::TRUCK) ? false : oncoming_status;
  return incorridor_debouncer_delay_map_[lane_id].first.oncoming_delay.RunDelay(
      oncoming_flag);
}

bool Obstacle::Run_rear_delay(const bool rear_flag) {
  return delay_rear_.RunDelay(rear_flag);
}

void Obstacle::Run_moving_delay() {
  delayed_motion_ = moving_delay_.RunDelay(
      motion_property_ == ObstacleMotionProperty::MOVING ||
      motion_property_ == ObstacleMotionProperty::MOVING_SLOWLY);
}

void Obstacle::Run_is_riding_marker_delay(const ObjectLocation &obj_loc,
                                          float lane_width) {
  if (obj_loc.in_corridor) {
    is_riding_on_lane_marker_ = riding_delay_.RunDelay(false);
  } else {
    static bool is_front = true;
    if (is_front) {
      if (obj_loc.reference_point2corridor_frenet.s() < 0.) {
        is_front = false;
        riding_delay_.SetRiseDelayTime(2.0);
      }
    } else {
      if (obj_loc.reference_point2corridor_frenet.s() > 0.) {
        is_front = true;
        riding_delay_.SetRiseDelayTime(1.0);
      }
    }
    float lateral_velocity = obj_loc.reference_point2corridor_frenet.dl();
    float l = obj_loc.reference_point2corridor_frenet.l();
    float vehedge2refline = std::abs(l) - 0.5 * width_;
    is_riding_on_lane_marker_ =
        riding_delay_.RunDelay(std::abs(lateral_velocity) < 0.3 &&
                               (vehedge2refline < lane_width * 0.5 + 0.1) &&
                               (vehedge2refline > lane_width * 0.5 - 0.8));
  }
}

void Obstacle::OvertakingJudge(const ObjectLocation &obj_loc,
                               float ref_velocity) {
  float delta_velocity =
      obj_loc.reference_point2corridor_frenet.ds() - ref_velocity;
  if (is_overtaking_) {
    is_overtaking_ = overtaking_delay_.RunDelay(delta_velocity > -0.5);
  } else {
    is_overtaking_ = overtaking_delay_.RunDelay(delta_velocity > 1.0);
  }
}

void Obstacle::Run_is_parallel_delay(const ObjectLocation &obj_loc) {
  if (obj_loc.in_corridor) {
    is_parallel_obstacle_ = parallel_delay_.RunDelay(false);
  } else {
    // float delta_velocity =
    //     rm::VehicleStateProvider::Instance()->lon_velocity() -
    //     obj_loc.reference_point2corridor_frenet.ds();
    float half_length = 0.5 * length_;
    float front_to_object_back = obj_loc.reference_point2corridor_frenet.s() -
                                 half_length - rm::DISREARAXEL2FRONT;
    float back_to_object_front = obj_loc.reference_point2corridor_frenet.s() +
                                 half_length + rm::DISREARAXEL2BACK;
    bool is_overlapped =
        front_to_object_back < 5.0F && back_to_object_front > -5.0F;
    is_parallel_obstacle_ = parallel_delay_.RunDelay(is_overlapped);
  }
}

bool Obstacle::is_riding_on_lane_marker() const {
  return is_riding_on_lane_marker_;
}

bool Obstacle::is_overtaking() const { return is_overtaking_; }

bool Obstacle::is_parallel_obstacle() const { return is_parallel_obstacle_; }

bool Obstacle::Generate_virtual_corridor() {
  // return false if conditions are not  fullfilled
  if (virtual_corridor_generation_plausibility_ &&
      (historical_traj_.size() >= 15)) {
    std::vector<std::pair<ReferencePoint, double>> reference_traj_points;
    for (auto &&point : historical_traj_) {
      if (point.x() >= -5.0) {
        reference_traj_points.emplace_back(
            std::make_pair(ReferencePoint(point), 1));
      }
    }
    virtual_corridor_.set_feature_tag(CorridorFeatureTag::CARFOLLOW);
    virtual_corridor_.set_corridor_id(id_); // same as object id
    virtual_corridor_.ComputeCenterLine(reference_traj_points);
    double virtual_s =
        virtual_corridor_.mutable_center_line()->ComputeArcLengthbySimpson(
            0.0, center_.x());
    ObjectLocation virtual_object_location(
        id_, virtual_s, speed_.Length(), 0.0f, 0.0f, 0.0f, 0.0f, true,
        rm::ObstacleType::UNKNOWN, center_.x(), center_.y(), width_,
        length_); // uint32_t id, float s, float ds, float dds, float l, float
                  // dl, float ddl, bool in_flag
    if (!virtual_corridor_.mutable_object_location()->empty()) {
      // object already exists, delete the obsolete info, and push the new one
      virtual_corridor_.DeleteObjectLocation();
    }
    virtual_corridor_.AddObjectLocation(virtual_object_location);
    return true;
  } else {
    return false;
  }
}

bool Obstacle::Generate_virtual_corridor(
    const std::pair<Boundary, Boundary> *lane_info_boundary) {
  // return false if conditions are not  fullfilled
  // this function generates the car follow corridor when lane presents but
  // corridor does not exist, so we need to check if the generated corridor has
  // intersection with boundaries on both sides
  if (virtual_corridor_generation_plausibility_ &&
      (historical_traj_.size() >= 15)) {
    std::vector<std::pair<ReferencePoint, double>> reference_traj_points;
    std::vector<double> min_distance2boundaries;
    bool is_left_marker_empty =
        lane_info_boundary->first.boundary_segment_list().empty();
    bool is_right_marker_empty =
        lane_info_boundary->second.boundary_segment_list().empty();
    PolynomLine left_curve =
        is_left_marker_empty ? PolynomLine({5.0, 0.0, 0.0, 0.0}, 0.0, 50.0)
                             : lane_info_boundary->first.boundary_segment_list()
                                   .front()
                                   .geometry.curve();
    PolynomLine right_curve =
        is_right_marker_empty
            ? PolynomLine({-5.0, 0.0, 0.0, 0.0}, 0.0, 50.0)
            : lane_info_boundary->second.boundary_segment_list()
                  .front()
                  .geometry.curve();
    for (auto &&point : historical_traj_) {
      // reference_traj_points.emplace_back(std::make_pair(ReferencePoint(point),
      // 1));
      if (point.x() >= -5.0) {
        // compute the distance between point and line boundary
        reference_traj_points.emplace_back(
            std::make_pair(ReferencePoint(point), 1));
        // both dis_trajpoint2leboundpoint and dis_trajpoint2riboundpoint shall
        // be greater than 0 in order to generate corridor that does not
        // intersect boundaries
        double dis_trajpoint2leboundpoint, dis_trajpoint2riboundpoint;
        if (is_left_marker_empty) {
          dis_trajpoint2leboundpoint = 10.0;
        } else {
          if (point.x() > left_curve.end_range() - 4.0F) {
            dis_trajpoint2leboundpoint = 10.0;
          } else {
            dis_trajpoint2leboundpoint =
                left_curve.Evaluate(0, point.x()) - point.y();
          }
        }
        if (is_right_marker_empty) {
          dis_trajpoint2riboundpoint = 10.0;
        } else {
          if (point.x() > right_curve.end_range() - 4.0) {
            dis_trajpoint2riboundpoint = 10.0;
          } else {
            dis_trajpoint2riboundpoint =
                point.y() - right_curve.Evaluate(0, point.x());
          }
        }
        if (dis_trajpoint2leboundpoint < 0.4F ||
            dis_trajpoint2riboundpoint < 0.4F) {
          // the generated corridor is not traversable
          // DLOG(INFO) << "the generated corridor is not traversable";
          return false;
        } else {
          min_distance2boundaries.push_back(
              std::min(dis_trajpoint2leboundpoint, dis_trajpoint2riboundpoint));
        }
      } else {
        // do nothing
      }
    }
    virtual_corridor_.set_feature_tag(CorridorFeatureTag::CARFOLLOW);
    virtual_corridor_.set_corridor_id(id_); // same as object id
    virtual_corridor_.ComputeCenterLine(reference_traj_points);
    double virtual_s =
        virtual_corridor_.mutable_center_line()->ComputeArcLengthbySimpson(
            0.0, center_.x());
    if (!min_distance2boundaries.empty()) {
      // if vector is not empty, take the minimal value and set virtual drivable
      // width to it
      std::vector<double>::iterator min_iter = std::min_element(
          min_distance2boundaries.begin(), min_distance2boundaries.end());
      virtual_corridor_.set_virtualdrivablewidth(*min_iter);
    } else {
      //  if there is no enough sample points, set to default value
      virtual_corridor_.set_virtualdrivablewidth(3.5f);
    }
    ObjectLocation virtual_object_location(
        id_, virtual_s, speed_.Length(), 0.0f, 0.0f, 0.0f, 0.0f, true,
        rm::ObstacleType::UNKNOWN, center_.x(), center_.y(), width_,
        length_); // uint32_t id, float s, float ds, float dds, float l, float
                  // dl, float ddl, bool in_flag
    if (!virtual_corridor_.mutable_object_location()->empty()) {
      // object already exists, delete the obsolete info, and push the new one
      virtual_corridor_.DeleteObjectLocation();
    }
    virtual_corridor_.AddObjectLocation(virtual_object_location);
    return true;
  } else {
    return false;
  }
}

void Obstacle::Check_virtual_corridor_generation_plausibility() {
  // pre-conditions check for obstacle virtual trajectory generation
  virtual_corridor_generation_plausibility_ =
      longdistance_debouncer_delay_.RunDelay(
          max_longdistance_debouncer_.RunDebouceLessTure(center_.x()) &&
          min_longdistance_debouncer_.RunDebouceMoreTure(center_.x())) &&
      (VehicleStateProvider::Instance()->lon_velocity() <= 16.67);
}

const Corridor &Obstacle::virtual_corridor() const { return virtual_corridor_; }

bool Obstacle::operator==(const uint32_t id) const { return id_ == id; }

void Obstacle::Bind_with_lane(uint8_t lane_id) {
  if (incorridor_debouncer_delay_map_.count(lane_id) == 0) // appears first time
  {
    // std::array<Debouncer, 2> debouncer_array = {Debouncer(), Debouncer()};
    // incorridor_debouncer_delay_map_[lane_id] = std::make_pair(
    //     std::make_pair(debouncer_array, Delay<bool>(0.16, 0.0, false)),
    //     true);
    incorridor_debouncer_delay_map_[lane_id] =
        std::make_pair(InCorridorDDStruct(Delay<bool>(0.16, 0.0, false),
                                          Delay<bool>(0.08, 0, false)),
                       true);
  } else // already exist
  {
    incorridor_debouncer_delay_map_[lane_id].second = true;
  }
}

void Obstacle::DeleteandReset_incorridor_map() {
  for (auto iter = incorridor_debouncer_delay_map_.begin();
       iter != incorridor_debouncer_delay_map_.end();) {
    if (iter->second.second == false) {
      // erase the element in the map if the binded lane no longer exists
      incorridor_debouncer_delay_map_.erase(iter++);
    } else {
      // else set the lane existence flag to false
      iter->second.second = false;
      iter++;
    }
  }
}

TargetType Obstacle::trplontype() const { return trplontype_; }

TargetType Obstacle::prv_trplontype() const { return prv_trplontype_; }

void Obstacle::set_trplontype(TargetType target_type) {
  trplontype_changed_ = true;
  prv_trplontype_ = trplontype_;
  trplontype_ = target_type;
  if (target_type == TargetType::FLNIV || target_type == TargetType::FLNIIV ||
      target_type == TargetType::SLNIV || target_type == TargetType::SLNIIV) {
    is_front_adjacent_target_ = true;
  }
}

void Obstacle::Run_in_corridor_long_distance_debouncer(
    double ego_velocity_kph) {
  if (classification_ == ObstacleType::PEDESTRIAN) {
    double lower_range = 3.5 * std::max(ego_velocity_kph * rm::KPH2MPS, 2.0);
    in_corridor_long_debouncer_.SetDebounce(lower_range, lower_range + 5.0);
    is_in_long_distance_range_ =
        in_corridor_long_debouncer_.RunDebouceLessTure(center_.x());
  } else {
    in_corridor_long_debouncer_.SetDebounce(
        in_corridor_lower_long_dis_ego_velocity_table_.Evaluate(
            ego_velocity_kph),
        in_corridor_upper_long_dis_ego_velocity_table_.Evaluate(
            ego_velocity_kph));
    is_in_long_distance_range_ = in_corridor_long_debouncer_.RunDebouceLessTure(
        center_.x() - 0.5 * length_);
    // take into account the length of objects
  }
}

bool Obstacle::is_in_long_distance_range() const {
  return is_in_long_distance_range_;
}

void Obstacle::reset_trplontype() {
  if (!trplontype_changed_) {
    prv_trplontype_ = trplontype_;
    trplontype_ = TargetType::OTHER;
  }
}

float Obstacle::existence_duration() const { return existence_duration_; }

void Obstacle::Update_existence_duration() {
  existence_duration_ =
      std::min(existence_duration_ + rm::ROADMODELCYCLEINS, 1000.0);
}

void Obstacle::set_prv_corridor_id(uint32_t prv_corridor_id) {
  prv_corridor_id_ = prv_corridor_id;
}

void Obstacle::set_is_already_assigned() { is_already_assgined_ = true; }

void Obstacle::unset_is_already_assigned() { is_already_assgined_ = false; }

bool Obstacle::is_already_assigned() const { return is_already_assgined_; }

void Obstacle::unset_prv_corridor_id() { prv_corridor_id_ = 0U; }

uint32_t Obstacle::prv_corridor_id() const { return prv_corridor_id_; }

void Obstacle::Add_duration_of_intermidiate_status() {
  duration_of_intermidiate_status_ += rm::ROADMODELCYCLEINS;
}

void Obstacle::Reset_duration_of_intermidiate_status() {
  duration_of_intermidiate_status_ = 0.0F;
}

float Obstacle::duration_of_intermidiate_status() {
  return duration_of_intermidiate_status_;
}

void Obstacle::Update_incorridor_map_key(uint8_t lane_id_last_cycle,
                                         uint8_t lane_id) {
  // return if current lane id is invalid
  if (lane_id == 0)
    return;
  if (incorridor_debouncer_delay_map_.count(lane_id_last_cycle) == 1) {
    // the lane id existed in the map
    incorridor_debouncer_delay_map_[lane_id] =
        incorridor_debouncer_delay_map_[lane_id_last_cycle];
    incorridor_debouncer_delay_map_.erase(lane_id_last_cycle);
  }
}

bool Obstacle::lane_change_intention() const { return lane_change_intention_; }

void Obstacle::Reset_lane_change_intention_detector() {
  lane_change_intention_detector_.Reset();
  lane_change_intention_ = false;
}

void Obstacle::Run_lane_change_intention_detector(
    const ObjectLocation &obj_loc) {
  bool is_left_obj =
      (trplontype_ == TargetType::FLNIV || trplontype_ == TargetType::FLNIIV);
  lane_change_intention_ =
      lane_change_intention_detector_.Run(obj_loc, is_left_obj);
  // DLOG(INFO) << PRINT(speed_.y());
}

bool Obstacle::is_front_adjacent_target() const {
  return is_front_adjacent_target_;
}

void Obstacle::Compute_prediction_factor_ego_yaw_rate() {
  width_factor_ego_yaw_rate_ = width_factor_ego_yaw_rate_table_.Evaluate(
      VehicleStateProvider::Instance()->yaw_rate());
  // DLOG(INFO) << PRINT(width_factor_ego_yaw_rate_);
}

void Obstacle::Reset_was_in_ego_corridor_assured_flfp() {
  was_in_ego_corridor_assured_flfp_.Reset();
}

void Obstacle::set_is_vulnerable_critical_obstacle(bool flag) {
  is_vulnerable_critical_obstacle_ = flag;
}

void Obstacle::
    Reset_vulnerable_object_critical_ego_corridor_debouncer_pair_pre() {
  vulnerable_object_critical_ego_corridor_debouncer_pair.lat_dis_debouncer
      .ResetDebounce();
  vulnerable_object_critical_ego_corridor_debouncer_pair.long_dis_debouncer
      .ResetDebounce();
  vulnerable_object_critical_ego_corridor_debouncer_pair.excuted_flag = false;
}

void Obstacle::
    Reset_vulnerable_object_critical_ego_corridor_debouncer_pair_post() {
  if (!vulnerable_object_critical_ego_corridor_debouncer_pair.excuted_flag) {
    // if vulnerable_object_critical_ego_corridor_debouncer_pair not being
    // excuted, reset debouncer
    vulnerable_object_critical_ego_corridor_debouncer_pair.lat_dis_debouncer
        .ResetDebounce();
    vulnerable_object_critical_ego_corridor_debouncer_pair.long_dis_debouncer
        .ResetDebounce();
  }
  vulnerable_object_critical_ego_corridor_debouncer_pair.excuted_flag = false;
}

bool Obstacle::is_vulnerable_critical_obstacle() const {
  return is_vulnerable_critical_obstacle_;
}

} // namespace rm
